########################################################################
# File:    task_imports_controller.rb                                  #
#          Hipposoft 2008                                              #
#                                                                      #
# Purpose: Import task definition data from external sources.          #
#                                                                      #
# History: 30-Mar-2008 (ADH): Created.                                 #
########################################################################

class TaskImportsController < ApplicationController

  require 'zlib'
  require 'ostruct'
  require 'tempfile'
  require 'rexml/document'

  # Security.

  before_filter( :permitted? )

  # Set up the import view. If there is no task data, this will consist of
  # a file entry field and nothing else. If there is parsed file data (a
  # preliminary task list), then this is included too.

  def new
  end

  # Take the task data from the 'new' view form and 'create' an "import
  # session"; that is, create real Task objects based on the task list and
  # add them to the database, wrapped in a single transaction so that the
  # whole operation can be unwound in case of error.

  def create

    # Set up a new TaskImport session object and read the XML file details

    xmlfile = params[ :import ][ :xmlfile ]
    @import = TaskImport.new
    @import.project_id = @current_user.control_panel.project_id

    if ( xmlfile != '' )

      # The user selected a file to upload, so process it

      begin

        # We assume XML files always begin with "<" in the first byte and
        # if that's missing then it's GZip compressed. That's true in the
        # limited case of project files.

        byte = xmlfile.getc()
        xmlfile.rewind()

        xmlfile       = Zlib::GzipReader.new( xmlfile ) if ( byte != '<'[ 0 ] )
        xmldoc        = REXML::Document.new( xmlfile.read() )
        @import.tasks = get_tasks_from_xml( xmldoc )

        if ( @import.tasks.nil? or @import.tasks.empty? )
          flash[ :error  ] = 'No usable tasks were found in that file'
        else
          flash[ :notice ] = 'Tasks read successfully. Please choose items to import.'
        end

      rescue => error

        # REXML errors can be huge, including a full backtrace. It can cause
        # session cookie overflow and we don't want the user to see it. Cut
        # the message off at the first newline.

        lines = error.message.split("\n")
        flash[ :error  ] = "Failed to read file: #{ lines[ 0 ] }"
      end

      render( { :action => :new } )
      flash.delete( :error  )
      flash.delete( :notice )

    else

      # No file was specified. If there are no tasks either, complain.

      tasks = params[ :import ][ :tasks ]

      if ( tasks.nil? )
        flash[ :error ] = "Please choose a file before using the 'Analyse' button."
        render( { :action => :new } )
        flash.delete( :error  )
        return
      end

      # Compile the form submission's task list into something that the
      # TaskImport object understands.
      #
      # Since we'll rebuild the tasks array inside @import, we can render the
      # 'new' view again and have the same task list presented to the user in
      # case of error.

      @import.tasks = []
      to_import     = []

      # Due to the way the form is constructed, 'task' will be a 2-element
      # array where the first element contains a string version of the index
      # at which we should store the entry and the second element contains
      # the hash describing the task itself.

      tasks.each do | taskinfo |
        index  = taskinfo[ 0 ].to_i
        task   = taskinfo[ 1 ]
        struct = OpenStruct.new

        struct.title    = task[ :title    ]
        struct.level    = task[ :level    ]
        struct.code     = task[ :code     ]
        struct.duration = task[ :duration ]

        @import.tasks[ index ] = struct
        to_import[ index ] = struct if ( task[ :import ] == '1' )
      end

      to_import.compact!

      # The "import" button in the form causes token "import_selected" to be
      # set in the params hash. The "analyse" button causes nothing to be set.
      # If the user has clicked on the "analyse" button but we've reached this
      # point, then they didn't choose a new file yet *did* have a task list
      # available. That's strange, so raise an error.
      #
      # On the other hand, if the 'import' button *was* used but no tasks were
      # selected for error, raise a different error.

      if ( params[ :import ][ :import_selected ].nil? )
        flash[ :error ] = 'No new file was chosen for analysis. Please choose a file before using the "Analyse" button, or use the "Import" button to import tasks selected in the task list.'
      elsif ( to_import.empty? )
        flash[ :error ] = 'No tasks were selected for import. Please select at least one task and try again.'
      end

      # Bail out if we have errors to report.

      unless( flash[ :error ].nil? )
        render( { :action => :new } )
        flash.delete( :error  )
        return
      end

      # Right, good to go! Do the import.

      project_id = params[ :import ][ :project_id ]
      project    = nil

      begin
        Task.transaction do
          project = Project.find_by_id( project_id ) unless ( project_id.nil? or project_id.empty? )

          to_import.each do | source_task |
            destination_task          = Task.new
            destination_task.title    = source_task.title
            destination_task.code     = source_task.code
            destination_task.duration = source_task.duration
            destination_task.project  = project
            destination_task.save!
          end

          flash[ :notice ] = "#{ to_import.length } #{ to_import.length == 1 ? 'task' : 'tasks' } imported successfully."
          redirect_to( new_task_import_path() )
        end

      rescue => error
        flash[ :error ] = "Unable to import tasks: #{ error }"
        render( { :action => :new } )
        flash.delete( :error )

      end
    end
  end

private

  # Is the current action permitted?

  def permitted?
    appctrl_not_permitted() if( @current_user.restricted? )
  end

  # Obtain a task list from the given parsed XML data (a REXML document).

  def get_tasks_from_xml( doc )

    # Extract details of every task into a flat array

    tasks = []

    doc.each_element( 'Project/Tasks/Task' ) do | task |
      begin
        struct = OpenStruct.new

        struct.level = task.get_elements( 'OutlineLevel' )[ 0 ].text.to_i
        struct.tid   = task.get_elements( 'ID'           )[ 0 ].text.to_i
        struct.uid   = task.get_elements( 'UID'          )[ 0 ].text.to_i
        struct.title = task.get_elements( 'Name'         )[ 0 ].text

        tasks.push( struct )
      rescue
        # Ignore errors; they tend to indicate malformed tasks, or at least,
        # XML file task entries that we do not understand.
      end
    end

    # Sort the array by ID. By sorting the array this way, the order
    # order will match the task order displayed to the user in the
    # project editor software which generated the XML file.

    tasks = tasks.sort_by { | task | task.tid }

    # Step through the sorted tasks. Each time we find one where the
    # *next* task has an outline level greater than the current task,
    # then the current task MUST be a summary. Record its name and
    # blank out the task from the array. Otherwise, use whatever
    # summary name was most recently found (if any) as a name prefix.

    prefix = ''

    tasks.each_index do | index |
      task      = tasks[ index     ]
      next_task = tasks[ index + 1 ]

      if ( next_task and next_task.level > task.level )
        prefix         = task.title
        tasks[ index ] = nil
      else
        task.title = "#{ prefix }: #{ task.title }" unless prefix.empty?
      end
    end

    # Remove any 'nil' items we created above

    tasks.compact!

    # Now create a secondary array, where the UID of any given task is
    # the array index at which it can be found. This is just to make
    # looking up tasks by UID really easy, rather than faffing around
    # with "tasks.find { | task | task.uid = <whatever> }".

    uid_tasks = []

    tasks.each do | task |
      uid_tasks[ task.uid ] = task
    end

    # OK, now it's time to parse the assignments into some meaningful
    # array. These will become our timesheet system tasks. Assignments
    # which relate to empty elements in "uid_tasks" or which have zero
    # work are associated with tasks which are either summaries or
    # milestones. Ignore both types.

    real_tasks = []

    doc.each_element( 'Project/Assignments/Assignment' ) do | as |
      task_uid = as.get_elements( 'TaskUID' )[ 0 ].text.to_i
      task = uid_tasks[ task_uid ]
      next if ( task.nil? )

      work = as.get_elements( 'Work' )[ 0 ].text

      # Parse the "Work" string: "PT<num>H<num>M<num>S", but with some
      # leniency to allow any data before or after the H/M/S stuff.

      strs = work.scan(/.*?(\d+)H(\d+)M(\d+)S.*?/).flatten
      hours, mins, secs = strs.map { | str | str.to_i }

      next if ( hours == 0 and mins == 0 and secs == 0 )

      # Woohoo, real task!

      today         = Date.current
      task.code     = "XML-%04d%02d%02d-#{ task.tid }" % [ today.year, today.month, today.day ]
      task.duration = ( ( ( hours * 3600 ) + ( mins * 60 ) + secs ) / 3600 ).precision( 2 )

      real_tasks.push( task )
    end

    return real_tasks
  end
end
